# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import dataclasses
import enum
from typing import Any, Dict


class _BaseModel:
    def __iter__(self):
        for attr, value in self.__dict__.items():
            yield attr, value

    def to_dict(self) -> Dict[str, Any]:
        model_dict = {**dataclasses.asdict(self)}

        for k, v in model_dict.items():
            if isinstance(v, enum.Enum):
                model_dict[k] = repr(v)
        return model_dict
